///////////////////////////////////////////////////////////////////////////////
//
//  Die Uhr (c) 2006 by Kay Bruns
//
//  I wrote this program to have a neat looking analog desktop clock.
//  
//  Most clocks are Payware so I spend one day to make a freeware Clock
///////////////////////////////////////////////////////////////////////////////

//We need WinXP!
#define _WIN32_WINNT 0x0501
#define WINVER       0x0501
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <Shobjidl.h>
#include <GdiPlus.h>
#pragma comment(lib,"gdiplus.lib")
using namespace Gdiplus;
#include <WinGDI.h>
#pragma comment(lib,"msimg32.lib")
#include <shlwapi.h>
#pragma comment(lib, "shlwapi.lib")
#include "Resource.h"

///////////////////////////////////////////////////////////////////////////////
// Global Stuff
///////////////////////////////////////////////////////////////////////////////
#define ClassName L"DieUhrWindowClass"
#define RgKey     L"Software\\Kay Bruns\\DieUhr"
#define AutoRunKey L"Software\\Microsoft\\Windows\\CurrentVersion\\Run"

HINSTANCE g_hInst;                //Guess!
int g_XPos=0;                     //initial X Position for the clock
int g_YPos=0;                     //initial Y Position for the clock
int g_Scale = 100;                //Scale for the Clock in %
int g_Alpha = 100;                //Clock Alpha 0%=Transparent, 100%=Opaque
TCHAR g_SkinDir[MAX_PATH]={0};    //Directory for the current Skin files
DWORD g_LastSecond=-1;            //Last displayed 12th part of a Second
DWORD g_LastMinute=-1;            //Last displayed 12th part of a Minute
Bitmap* g_ClockBM=0;              //Clock Background Bitmap "main.png"
Bitmap* g_HourBM=0;               //Clock Hours Hand Bitmap "hour.bmp"
Bitmap* g_MinBM=0;                //Clock Minutes Hand Bitmap "min.bmp"
Bitmap* g_SecBM=0;                //Clock Seconds Hand Bitmap "sec.bmp"
Bitmap* g_GlassBM=0;              //Clock Glass Bitmap
Bitmap* g_DrawBM=0;               //Memory Bitmap for drawing
CachedBitmap* g_ClockCache=0;     //Cached Bitmap for Back+Hour+Min
Graphics* g_gr=0;                 //Seconds Graphics object
SIZE g_Size={0};                  //The Background Bitmap size
bool g_ShowSeconds=TRUE;          //Show "sec.png"
bool g_ShowGlass=TRUE;            //Overlay "attach.png"
bool g_Transparent=false;         //WS_EX_TRANSPARENT: Pass Mouse through
                                  //You must use Alt+TAB to activate the Clock!
bool g_TopMost=false;             //WS_EX_TOPMOST: Always on Top

double pi=3.141592654;            //a Shame C++ does not have this!
double g_HndScale=0;              //"Alwact clock" Skins Hands are 5 times 
                                  //larger here's a scaling factor to fix this
HDC g_MemDC=0;                    //The Memory DC for UpdateLayeredWindow
HBITMAP g_hbm=0;                  //g_DrawBM's HDC usually selected into MemDC

/////////////////////////////////////////////////////////////////////////////
// Registry stuff                                      I'll not explain this!
/////////////////////////////////////////////////////////////////////////////
BOOL GetRegAny(HKEY HK,LPCTSTR SubKey,LPCTSTR ValName,DWORD Type,BYTE* RetVal,DWORD nBytes)
{
  HKEY Key;
  if (RegOpenKeyEx(HK,SubKey,0,KEY_READ,&Key)==ERROR_SUCCESS)
  {
    DWORD dwType=Type;
    DWORD l=nBytes;
    BOOL bRet=(RegQueryValueEx(Key,ValName,NULL,&dwType,RetVal,&l)==ERROR_SUCCESS)
            &&(dwType==Type);
      RegCloseKey(Key);
    return bRet;
  }
  return FALSE;
}

BOOL SetRegAny(HKEY HK,LPCTSTR SubKey,LPCTSTR ValName,DWORD Type,BYTE* Data,DWORD nBytes)
{
  HKEY Key;
  BOOL bKey=FALSE;
  bKey=RegOpenKeyEx(HK,SubKey,0,KEY_WRITE,&Key)==ERROR_SUCCESS;
  if (!bKey)
    bKey=RegCreateKey(HK,SubKey,&Key)==ERROR_SUCCESS;
  if (bKey)
  {
    LONG l=RegSetValueEx(Key,ValName,0,Type,Data,nBytes);
    RegCloseKey(Key);
    return l==ERROR_SUCCESS;
  }
  return FALSE;
}

BOOL RegDelVal(HKEY HK,LPCTSTR SubKey,LPCTSTR ValName)
{
  HKEY Key;
  if (RegOpenKeyEx(HK,SubKey,0,KEY_WRITE,&Key)==ERROR_SUCCESS)
  {
    BOOL bRet=RegDeleteValue(Key,ValName)==ERROR_SUCCESS;
    RegCloseKey(Key);
    return bRet;
  }
  return FALSE;
}

DWORD GetRegInt(HKEY HK,LPCTSTR SubKey,LPCTSTR ValName,DWORD Default)
{
  DWORD RetVal=0;
  if (GetRegAny(HK,SubKey,ValName,REG_DWORD,(BYTE*)&RetVal,sizeof(RetVal)))
    return RetVal;
  return Default;
}

static TCHAR* GetRegStr(HKEY HK,LPCTSTR SubKey,LPCTSTR ValName,LPTSTR Default)
{
  static TCHAR Str[4096]={0};
  if (GetRegAny(HK,SubKey,ValName,REG_SZ,(BYTE*)&Str,sizeof(Str)/sizeof(TCHAR)-sizeof(TCHAR)))
    return Str;
  if (GetRegAny(HK,SubKey,ValName,REG_EXPAND_SZ,(BYTE*)&Str,sizeof(Str)/sizeof(TCHAR)-sizeof(TCHAR)))
    return Str;
  return Default;
}

BOOL SetRegInt(HKEY HK,LPCTSTR SubKey,LPCTSTR ValName,DWORD Value)
{
  return SetRegAny(HK,SubKey,ValName,REG_DWORD,(BYTE*)&Value,sizeof(DWORD));
}

BOOL SetRegStr(HKEY HK,LPCTSTR SubKey,LPCTSTR ValName,LPCTSTR Value)
{
  return SetRegAny(HK,SubKey,ValName,REG_SZ,(BYTE*)Value,wcslen(Value)*sizeof(TCHAR));
}

///////////////////////////////////////////////////////////////////////////////
// PNGFromResource: from CodeGuru.com
//
//  Load a PNG directly from a resource to a GDI+ Bitmap
///////////////////////////////////////////////////////////////////////////////
Bitmap* LoadPNGResource(LPCTSTR pName)
{
  HRSRC hResource = ::FindResource(g_hInst, pName, L"PNG_FILE");
  if (!hResource)
    return NULL;
  DWORD imageSize = ::SizeofResource(g_hInst, hResource);
  if (!imageSize)
    return NULL;
  const void* pResourceData = ::LockResource(::LoadResource(g_hInst, hResource));
  if (!pResourceData)
    return NULL;
  HANDLE hBuffer = ::GlobalAlloc(GMEM_MOVEABLE, imageSize);
  if (!hBuffer)
    return NULL;
  void* pBuffer = ::GlobalLock(hBuffer);
  if (pBuffer)
  {
    CopyMemory(pBuffer, pResourceData, imageSize);
    IStream* pStream = NULL;
    if (::CreateStreamOnHGlobal(hBuffer, FALSE, &pStream) == S_OK)
    {
      Bitmap* pBitmap = Bitmap::FromStream(pStream);
      pStream->Release();
      if (pBitmap)
      { 
        if (pBitmap->GetLastStatus() == Gdiplus::Ok)
        {
          ::GlobalUnlock(hBuffer);
          ::GlobalFree(hBuffer);
          return pBitmap;
        }
        delete pBitmap;
        pBitmap = NULL;
      }
    }
    ::GlobalUnlock(hBuffer);
  }
  ::GlobalFree(hBuffer);
  return NULL;
}

///////////////////////////////////////////////////////////////////////////////
//State Saving/Loading
///////////////////////////////////////////////////////////////////////////////

void SaveState(HWND hWnd)
{
  RECT wr={0};
  GetWindowRect(hWnd,&wr);
  SetRegInt(HKEY_CURRENT_USER,RgKey,L"PosX",wr.left);
  SetRegInt(HKEY_CURRENT_USER,RgKey,L"PosY",wr.top);
  SetRegInt(HKEY_CURRENT_USER,RgKey,L"Scale",g_Scale);
  SetRegInt(HKEY_CURRENT_USER,RgKey,L"Alpha",g_Alpha);
  SetRegStr(HKEY_CURRENT_USER,RgKey,L"Skin",g_SkinDir);
  SetRegInt(HKEY_CURRENT_USER,RgKey,L"Transparent",g_Transparent);
  SetRegInt(HKEY_CURRENT_USER,RgKey,L"TopMost",g_TopMost);
  SetRegInt(HKEY_CURRENT_USER,RgKey,L"ShowSeconds",g_ShowSeconds);
  SetRegInt(HKEY_CURRENT_USER,RgKey,L"ShowGlass",g_ShowGlass);
}

void LoadState()
{
  g_XPos =  GetRegInt(HKEY_CURRENT_USER,RgKey,L"PosX",150);
  g_YPos =  GetRegInt(HKEY_CURRENT_USER,RgKey,L"PosY",100);
  g_Scale = GetRegInt(HKEY_CURRENT_USER,RgKey,L"Scale",100);
  g_Alpha = GetRegInt(HKEY_CURRENT_USER,RgKey,L"Alpha",100);
  wcscpy(g_SkinDir,GetRegStr(HKEY_CURRENT_USER,RgKey,L"Skin",L""));
  g_Transparent=GetRegInt(HKEY_CURRENT_USER,RgKey,L"Transparent",0)!=0;
  g_TopMost=GetRegInt(HKEY_CURRENT_USER,RgKey,L"TopMost",0)!=0;
  g_ShowSeconds=GetRegInt(HKEY_CURRENT_USER,RgKey,L"ShowSeconds",1)!=0;
  g_ShowGlass=GetRegInt(HKEY_CURRENT_USER,RgKey,L"ShowGlass",1)!=0;
}

///////////////////////////////////////////////////////////////////////////////
// UpdateWindowStyle: Set TopMost ans Transparent styles
///////////////////////////////////////////////////////////////////////////////
void UpdateWindowStyle(HWND hWnd,BOOL bIgnoreFocus=FALSE)
{
	DWORD dwStyle = GetWindowLong(hWnd, GWL_EXSTYLE);
  if ( g_Transparent &&(bIgnoreFocus||(GetFocus()!=hWnd)))
    dwStyle |=WS_EX_TRANSPARENT;
  else
    dwStyle &=~WS_EX_TRANSPARENT;
  SetWindowLong(hWnd, GWL_EXSTYLE,dwStyle);
  SetWindowPos(hWnd,g_TopMost?HWND_TOPMOST:HWND_NOTOPMOST,0,0,0,0,
    SWP_NOSIZE|SWP_NOMOVE|SWP_NOACTIVATE|SWP_NOSENDCHANGING);
}

///////////////////////////////////////////////////////////////////////////////
// DeleteTaskBarIcon
//  I wanted DieUhr to appear in "Alt-Tab" Window but not in the Task Bar
//  Here's the code to do this
///////////////////////////////////////////////////////////////////////////////
void DeleteTaskBarIcon(HWND hWnd)
{
  //Note: CoInitialize and CoUninitialize are called in WinMain!
  ITaskbarList* pTaskbar=0;
  CoCreateInstance(CLSID_TaskbarList,0, CLSCTX_INPROC_SERVER,IID_ITaskbarList,(void**)&pTaskbar);
  pTaskbar->DeleteTab(hWnd);
}

///////////////////////////////////////////////////////////////////////////////
// Skin support
///////////////////////////////////////////////////////////////////////////////
void FreeSkin()
{
  //Delete all resources occupied by the skin Bitmaps
  //I hope I did not generate memory leaks here ;-)
  if (g_MemDC)
    g_hbm=(HBITMAP)SelectObject(g_MemDC,g_hbm);
  delete g_ClockCache;
  delete g_gr;
  delete g_ClockBM;
  delete g_HourBM;
  delete g_MinBM;
  delete g_SecBM;
  delete g_GlassBM;
  delete g_DrawBM;
  if (g_MemDC)
  {
    DeleteObject(g_hbm); 
    DeleteDC(g_MemDC);
  }
  g_gr=0;
  g_ClockCache=0;
  g_ClockBM=0;
  g_HourBM=0;
  g_MinBM=0;
  g_SecBM=0;
  g_GlassBM=0;
  g_DrawBM=0;
  g_MemDC=0;
  g_Size.cx=0;
  g_Size.cy=0;
}

void LoadSkin()
{
  FreeSkin();
  //Allocate all drawing stuff
  //Look in "<DieClockFolder>\Skins\<g_SkinDir>" for main.png
  TCHAR Path[MAX_PATH];
  GetModuleFileName(0,Path,MAX_PATH);
  PathRemoveFileSpec(Path);
  PathAppend(Path,L"Skins");
  PathAppend(Path,g_SkinDir);
  SetCurrentDirectory(Path);
  PathAppend(Path,L"main.png");
  if (PathFileExists(Path))
  {
    //Main.PNG found the remaining Bitmaps
    g_ClockBM=Bitmap::FromFile(L"main.png");
    if (g_ClockBM) //Just to be sure...
    {
      //g_DrawBM is 2x size of the background to have 200% Scale capability
      g_DrawBM=new Bitmap(g_ClockBM->GetWidth()*2,g_ClockBM->GetHeight()*2);
      g_HourBM=Bitmap::FromFile(L"hour.png");
      g_MinBM=Bitmap::FromFile(L"min.png");
      g_SecBM=Bitmap::FromFile(L"sec.png");
      g_GlassBM=Bitmap::FromFile(L"attach.png");
    }
  }
  if (!g_ClockBM) 
  {
    //Main.PNG not found or corrupt, load internal PNGs:
    g_SkinDir[0]=0;
    g_ClockBM=LoadPNGResource(L"MAIN");
    //g_DrawBM is 2x size of the background to have 200% Scale capability
    g_DrawBM=new Bitmap(g_ClockBM->GetWidth()*2,g_ClockBM->GetHeight()*2);
    g_HourBM=LoadPNGResource(L"HOUR");
    g_MinBM=LoadPNGResource(L"MIN");
    g_SecBM=LoadPNGResource(L"SEC");
    g_GlassBM=LoadPNGResource(L"attach");
  }
  g_Size.cx=g_ClockBM->GetWidth();
  g_Size.cy=g_ClockBM->GetHeight();
  //We use a MemDC for drawing and have the Bitmap permanently selected
  g_DrawBM->GetHBITMAP(0,&g_hbm);
  g_MemDC=CreateCompatibleDC(0);
  //Create main Graphics object
  g_hbm=(HBITMAP)SelectObject(g_MemDC,g_hbm);
  g_gr= new Graphics(g_MemDC);
  //make it draw smooth
  g_gr->SetInterpolationMode(InterpolationModeHighQualityBicubic);
  //create the cached bitmap of the background
  g_ClockCache=new CachedBitmap(g_DrawBM,g_gr);
  //reset time ticks
  g_LastSecond=-1;
  g_LastMinute=-1;
  //Scale clock hands: to be able to use Alwact Clock skins we must scale the
  //Hands by 0.2 if the Hands are bigger than the clock background
  double dXmax=0;
  g_HndScale=1.0;
  if(g_HourBM)
    dXmax=max(dXmax,g_HourBM->GetWidth());
  if(g_SecBM)
    dXmax=max(dXmax,g_SecBM->GetWidth());
  if(g_MinBM)
    dXmax=max(dXmax,g_MinBM->GetWidth());
  if((double)g_Size.cx*2.0/3.0<dXmax*g_HndScale)
    g_HndScale=0.2;
}

///////////////////////////////////////////////////////////////////////////////
// Popup Menu stuff
///////////////////////////////////////////////////////////////////////////////

void AddMenu(HMENU hMenu,int Percent,int BaseID,int CurPercent,bool bInvert)
{
  //little helper to create the % Menus
  TCHAR s[15];
  swprintf(s,L"%d%%",(bInvert?(100-Percent):Percent));
  AppendMenu(hMenu,(CurPercent==Percent)?MF_CHECKED:0,BaseID+Percent,s);
}

void DisplayMenu(HWND hWnd)
{
  POINT pt;
  GetCursorPos(&pt);
  //Main Menu
  HMENU hMenu=CreatePopupMenu();
  //Transparency menu
  HMENU hAlphaMenu=CreatePopupMenu();
  //Scale menu
  HMENU hScaleMenu=CreatePopupMenu();
  //Skin folder menu
  HMENU hSkinMenu=CreatePopupMenu();
  //Fill the Transparency submenu
  AddMenu(hAlphaMenu,100,100,g_Alpha,1);
  AddMenu(hAlphaMenu,85,100,g_Alpha,1);
  AddMenu(hAlphaMenu,80,100,g_Alpha,1);
  AddMenu(hAlphaMenu,75,100,g_Alpha,1);
  AddMenu(hAlphaMenu,68,100,g_Alpha,1);
  AddMenu(hAlphaMenu,60,100,g_Alpha,1);
  AddMenu(hAlphaMenu,50,100,g_Alpha,1);
  AddMenu(hAlphaMenu,37,100,g_Alpha,1);
  AddMenu(hAlphaMenu,20,100,g_Alpha,1);
  //Fill the Scale submenu
  AddMenu(hScaleMenu,20,200,g_Scale,0);
  AddMenu(hScaleMenu,25,200,g_Scale,0);
  AddMenu(hScaleMenu,32,200,g_Scale,0);
  AddMenu(hScaleMenu,40,200,g_Scale,0);
  AddMenu(hScaleMenu,50,200,g_Scale,0);
  AddMenu(hScaleMenu,63,200,g_Scale,0);
  AddMenu(hScaleMenu,80,200,g_Scale,0);
  AppendMenu(hScaleMenu,MF_SEPARATOR,-1,0);
  AddMenu(hScaleMenu,100,200,g_Scale,0);
  AppendMenu(hScaleMenu,MF_SEPARATOR,-1,0);
  AddMenu(hScaleMenu,125,200,g_Scale,0);
  AddMenu(hScaleMenu,160,200,g_Scale,0);
  AddMenu(hScaleMenu,200,200,g_Scale,0);
  //Fill the Skins submenu
  //The internal Skin is at Menu ID 499
  AppendMenu(hSkinMenu,(g_SkinDir[0]==0)?MF_CHECKED:0,499,L"Internal");
  AppendMenu(hSkinMenu,MF_SEPARATOR,-1,0);
  //The skins in the folder are at Menu IDs starting from 500
  int i=1;
  //Enumerate Skin SubDirectories
  TCHAR Path[MAX_PATH];
  TCHAR Dir[MAX_PATH];
  GetModuleFileName(0,Path,MAX_PATH);
  PathRemoveFileSpec(Path);
  wcscpy(Dir,Path);
  PathAppend(Path,L"Skins\\*.*");
  PathAppend(Dir,L"Skins");
  WIN32_FIND_DATA fd={0};
  HANDLE hFind=FindFirstFile(Path,&fd);
  if (hFind != INVALID_HANDLE_VALUE)
  {
    do
    {
      wcscpy(Path,Dir);
      PathAppend(Path,fd.cFileName);
      if ( PathIsDirectory(Path)
        &&(wcscmp(fd.cFileName,L".")!=0)
        &&(wcscmp(fd.cFileName,L"..")!=0))
      {
        //We found a path in the "Skins" Folder test if "main.png" is inside
        PathAppend(Path,L"main.png");
        if (PathFileExists(Path))
        {
          //We found a Skin subfolder, add it to the Skins submenu
          AppendMenu(hSkinMenu,(wcscmp(g_SkinDir,fd.cFileName)==0)?MF_CHECKED:0,500+i,fd.cFileName);
          i++;
        }
      }
    }while (FindNextFile(hFind,&fd));
  }
  FindClose(hFind);
  //Fill the main menu
  AppendMenu(hMenu,MF_POPUP,(DWORD)hScaleMenu,L"Scale");
  AppendMenu(hMenu,MF_POPUP,(DWORD)hAlphaMenu,L"Transparency");
  AppendMenu(hMenu,MF_SEPARATOR,-1,0);
  AppendMenu(hMenu,MF_POPUP,(DWORD)hSkinMenu,L"Skin");
  AppendMenu(hMenu,MF_SEPARATOR,-1,0);
  AppendMenu(hMenu,g_Transparent?MF_CHECKED:0,51,L"Mouse Pass Through");
  AppendMenu(hMenu,g_TopMost?MF_CHECKED:0,52,L"Always on Top");
  AppendMenu(hMenu,MF_SEPARATOR,-1,0);
  AppendMenu(hMenu,(g_ShowGlass?MF_CHECKED:0)|(g_GlassBM?0:MF_DISABLED),53,L"Glass");
  AppendMenu(hMenu,(g_ShowSeconds?MF_CHECKED:0)|(g_SecBM?0:MF_DISABLED),54,L"Seconds");
  AppendMenu(hMenu,MF_SEPARATOR,-1,0);
  AppendMenu(hMenu,0,50,L"Exit");
  //Show the popup
  int MenuID=TrackPopupMenu(hMenu,TPM_RETURNCMD|TPM_NONOTIFY,pt.x+1,pt.y+1,0,hWnd,NULL);
  //check the results
  if (MenuID==50)
    PostQuitMessage(0);
  else if (MenuID==51)
  {
    g_Transparent=!g_Transparent;
    //Transparency is a tricky toy, warn the user:
    if (g_Transparent)
      MessageBox(hWnd,
      L"NOTE! You can deactivate \"Mouse Pass Through\"\n"
      L"mode only by selecting Die Uhr with ALT+TAB!",L"Die Uhr",MB_ICONINFORMATION);
    UpdateWindowStyle(hWnd);
  }else if (MenuID==52)
  {
    g_TopMost=!g_TopMost;
    UpdateWindowStyle(hWnd);
  }else if (MenuID==53)
  {
    g_ShowGlass=!g_ShowGlass;
    g_LastMinute=-1;
  }else if (MenuID==54)
  {
    g_ShowSeconds=!g_ShowSeconds;
    g_LastMinute=-1;
  }else if ((MenuID>=110)&&(MenuID<=200))
    //New Alpha
    g_Alpha=MenuID-100;
  else if ((MenuID>=200)&&(MenuID<=400))
  {
    //new Scale
    g_Scale=MenuID-200;
    g_LastMinute=-1;
  }else if (MenuID==499)
  {
    //internal Skin
    g_SkinDir[0]=0;
    LoadSkin();
  }else if (MenuID>=500)
  {
    //new Skin
    GetMenuString(hSkinMenu,MenuID,g_SkinDir,MAX_PATH,MF_BYCOMMAND);
    LoadSkin();
  }
  //Clean up
  DestroyMenu(hMenu);
  DestroyMenu(hAlphaMenu);
  DestroyMenu(hScaleMenu);
};

///////////////////////////////////////////////////////////////////////////////
// Hand Drawing
///////////////////////////////////////////////////////////////////////////////

//The rotation center of the hands is at 1/2 dy and 1/3 dx
//Size of the Hand Bitmap parts left=1/3 right=2/3 bottom=1/2
#define HndL 1.0/3.0
#define HndR 2.0/3.0
#define HndT 1.0/2.0

//Angle 0..1.0 for 0..360
void DrawHand(double Angle,Graphics* gr,Bitmap* Hnd)
{
  //if Angle is Exactly 0, 0.25, 0.5 or 0.75 we get an OutOfMemory Error in
  //DrawImage below, so we adjust Angle by 0.09 to avoid this
  if (Angle*4 == (double)((int)(Angle*4)))
    Angle+=1/4096.0;
  //prescale Angle
  Angle = Angle*2.0*pi-pi/2;//Hands point to the left -pi/2 makes them point up
  //Here we calculate the three destination points for top left, top right and 
  //bottom left of the bitmap. It is hard to describe as Text, so please take a 
  //pen and make a drawing to see what's done here
  double cosa = cos(Angle);  //X-Part of the rotated Hand
  double sina = sin(Angle);  //Y-Part of the rotated Hand
  //Rotation center in the background
  double x0=(double)g_Size.cx*g_Scale/200.0; 
  double y0=(double)g_Size.cy*g_Scale/200.0;
  //Width and Height of scaled non rotated hand
  double dx=(double)Hnd->GetWidth()*g_Scale/100.0*g_HndScale;
  double dy=(double)Hnd->GetHeight()*g_Scale/100.0*g_HndScale;
  //The three Destination points scaled shifted and rotated
  PointF DstPts[]={
    //Upper Left
    PointF((float)(x0 - dx*HndL*cosa + dy*HndT*sina), (float)(y0 - dx*HndL*sina - dy*HndT*cosa)),
    //Upper Right
    PointF((float)(x0 + dx*HndR*cosa + dy*HndT*sina), (float)(y0 + dx*HndR*sina - dy*HndT*cosa)),
    //Lower left
    PointF((float)(x0 - dx*HndL*cosa - dy*HndT*sina), (float)(y0 - dx*HndL*sina + dy*HndT*cosa))
  };
  //Draw the Hand an the background
  Status st=gr->DrawImage(Hnd,DstPts,3);
}

///////////////////////////////////////////////////////////////////////////////
// Clock Update Procedure
///////////////////////////////////////////////////////////////////////////////
void UpdateTime(HWND hWnd)
{
  SYSTEMTIME st;
  GetLocalTime(&st);
  //Seconds are updated at 0.5 == 1/12 seconds
  if (g_LastSecond!=(DWORD)st.wSecond*12+3*(DWORD)st.wMilliseconds/250)
  {
    g_LastSecond=(DWORD)st.wSecond*12+3*(DWORD)st.wMilliseconds/250;
    //Hands are updated at 0.5 == 1/12 Minutes
    BOOL bUpd=FALSE;
    if ( (g_HourBM || g_MinBM) 
      && (g_LastMinute!=(DWORD)st.wMinute*12+(DWORD)st.wSecond/5) )
    {
      //Draw Hour and Minute to a cached Bitmap since it has only to be drawn 
      //once in 60 ticks
      bUpd=TRUE;
      g_LastMinute=(DWORD)st.wMinute*12+(DWORD)st.wSecond/5;
      //use a temp Graphics object
      Graphics gr1(g_DrawBM);
      gr1.SetInterpolationMode(InterpolationModeHighQualityBicubic);
      gr1.Clear(Color(0,0,0,0));
      gr1.DrawImage(g_ClockBM,0,0,g_Size.cx*g_Scale/100,g_Size.cy*g_Scale/100);
      //Only draw if we need to!
      if (g_HourBM)
        DrawHand(st.wHour/12.0+st.wMinute/720.0,&gr1,g_HourBM);
      //Only draw if we need to!
      if (g_MinBM)
        DrawHand(g_LastMinute/720.0,&gr1,g_MinBM);
      //store the resulting Bitmap to a chached bitmap
      delete g_ClockCache;
      g_ClockCache=new CachedBitmap(g_DrawBM,g_gr);
    }
    bUpd |=((g_SecBM!=0)&& g_ShowSeconds) || ((g_GlassBM!=0)&& g_ShowGlass);
    //Only draw if we need to!
    if (bUpd)
    {
      //Erase background and draw cached Bk/Hour/Min bitmap
      g_gr->Clear(Color(0,0,0,0));
      g_gr->DrawCachedBitmap(g_ClockCache,0,0);
      //Draw seconds if needed
      if (g_SecBM && g_ShowSeconds)
        DrawHand(g_LastSecond/720.0,g_gr,g_SecBM);
      //Draw Glass if needed
      if (g_GlassBM && g_ShowGlass)
      {
        //Fit the Glass to the center of the background
        double w=g_GlassBM->GetWidth()*g_Scale/100.0;
        double h=g_GlassBM->GetHeight()*g_Scale/100.0;
        double x0=(g_Size.cx*g_Scale/100.0-w)/2;
        double y0=(g_Size.cy*g_Scale/100.0-h)/2;
        g_gr->DrawImage(g_GlassBM,(REAL)x0,(REAL)y0,(REAL)w,(REAL)h);
      }
      //Update Window
      POINT ptBMP={0,0};
      BLENDFUNCTION bf = { AC_SRC_OVER, 0, g_Alpha*255/100, AC_SRC_ALPHA };
      SIZE WndSize={g_Size.cx*g_Scale/100,g_Size.cy*g_Scale/100};
      UpdateLayeredWindow(hWnd,0,0,&WndSize,g_MemDC,&ptBMP,0,&bf,ULW_ALPHA);
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
// Windows Procedure
///////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK WndProc(HWND hWnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
  switch (uMsg)
  {
  //Make Windows move the clock with a left click:
  case WM_NCHITTEST:
    return HTCAPTION;
#ifdef _DEBUG
  //I love pressing ESC when debugging Apps ;-)
  case WM_KEYDOWN:
    if (wParam==VK_ESCAPE)
      PostQuitMessage(0);
    return DefWindowProc(hWnd,uMsg,wParam,lParam);
#endif _DEBU
  //Context menu stuff
  case WM_CONTEXTMENU: 
  case WM_NCRBUTTONUP: 
    DisplayMenu(hWnd);
    break;
  //Post WM_QUIT on Alt+F4
  case WM_CLOSE: 
    PostQuitMessage(0);
    break; 
  //Store settings on WM_ENDSESSION
  case WM_ENDSESSION:
    //Die Uhr starts automatically after you log on if you did not close it 
    //before shutting down windows
    {
      TCHAR Path[MAX_PATH];
      GetModuleFileName(0,Path,MAX_PATH);
      SetRegStr(HKEY_CURRENT_USER,AutoRunKey,L"Die Uhr",Path);
    }
    FreeSkin();
    SaveState(hWnd);
    exit(0);
    break;
  //WM_TIMER is used to update the clock if nothing else happens
  case WM_TIMER:
    KillTimer(hWnd,wParam);
    UpdateTime(hWnd);
    SetTimer(hWnd,1,20,0);
    break;
  case WM_SETFOCUS:
    UpdateWindowStyle(hWnd);
    return DefWindowProc(hWnd,uMsg,wParam,lParam);
  //Shell32 creates a Taskbar Button if we get the focus
  //The Button is killed after the focus is gone
  case WM_KILLFOCUS:
    DeleteTaskBarIcon(hWnd);
    //DieUhr still has the Focus but it'll loose it soon
    //UpdateWindowStyle is forced to ignore the focus:
    UpdateWindowStyle(hWnd,true);
    return DefWindowProc(hWnd,uMsg,wParam,lParam);
  default: 
    //Check it the time's up to draw the clock:
    UpdateTime(hWnd);
    return DefWindowProc(hWnd,uMsg,wParam,lParam);
  }
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
//
//  Main
//
///////////////////////////////////////////////////////////////////////////////
int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
  //Singleton:
  CreateMutex(NULL,0,L"DIE_UHR_GEHT_SCHON");
  if (GetLastError()==ERROR_ALREADY_EXISTS)
    return 0;
  //Die Uhr is a Toy so adjust thread and process priorities
  SetPriorityClass(GetCurrentProcess(),IDLE_PRIORITY_CLASS);
  SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_IDLE);
  //Turn off AutoRun
  RegDelVal(HKEY_CURRENT_USER,AutoRunKey,L"Die Uhr");
  //Initialize PI with 64Bit accuracy ;)
  __asm
  {
    finit
    fldpi
    fstp pi
  };
  g_hInst=hInstance;
  //initialize COM
  CoInitialize(0);
  //initialize GDI+
  Gdiplus::GdiplusStartupInput gdiplusStartupInput;
  ULONG_PTR gdiplusToken;
  GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
  //Load Size, transparency, position and skin
  LoadState();
  LoadSkin();
  //Create Window
  WNDCLASS WndClass={0};
  WndClass.hCursor      = LoadCursor(0,IDC_ARROW);
  WndClass.lpszClassName= ClassName;
  WndClass.hInstance    = g_hInst;
  WndClass.style        = CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW;
  WndClass.lpfnWndProc  = WndProc;
  RegisterClass(&WndClass);
  HWND hWnd=CreateWindowEx(WS_EX_LAYERED,ClassName,L"Die Uhr",WS_POPUP|WS_VISIBLE,
    g_XPos,g_YPos,g_Size.cx*g_Scale/100,g_Size.cy*g_Scale/100,0,0,g_hInst,NULL);
  //Set Icons
  SendMessage(hWnd,WM_SETICON,ICON_BIG,(DWORD)LoadImage(g_hInst,MAKEINTRESOURCE(IDI_DIEUHR),IMAGE_ICON,32,32,LR_VGACOLOR));
  SendMessage(hWnd,WM_SETICON,ICON_SMALL,(DWORD)LoadImage(g_hInst,MAKEINTRESOURCE(IDI_DIEUHR),IMAGE_ICON,16,16,LR_VGACOLOR));
  //Set TopMost and Transparent styles
  UpdateWindowStyle(hWnd);
  DeleteTaskBarIcon(hWnd);
  //Run DieUhr
  SetTimer(hWnd,1,10,0);
  MSG Msg;
  while (GetMessage(&Msg,0,0,0))
  {
    TranslateMessage(&Msg);
    DispatchMessage(&Msg);
  }
  //Clean up
  SaveState(hWnd);
  DestroyWindow(hWnd);
  UnregisterClass(ClassName,g_hInst);
  FreeSkin();
  GdiplusShutdown(gdiplusToken);
  CoUninitialize();
	return 0;
}
  